# Copyright (C) 2021 Matthias Nadig

import numpy as np


class BoundFormatHandler:
    """
    A class that serves as interface to the actual array containing the bounds
    """

    def __init__(self, n_dims, shape):
        self._n_dims = n_dims
        self._shape = shape

    def __getitem__(self, key):
        bounds_extracted = self._getitem(key)
        if bounds_extracted.get_n_dims() != self.get_n_dims() or \
            len(bounds_extracted.get_shape()) == 0:
            raise RuntimeError('Invalid access of items in {}-dimensional bounds of shape {} using key: {}'.format(
                self.get_n_dims(), self.get_shape(), key
            ))
        return bounds_extracted

    def require_se(self):
        return self._require_se()

    def require_cw(self):
        return self._require_cw()

    def require_sw(self):
        return self._require_sw()

    def require_secw(self):
        return self._require_secw()

    def require_start(self):
        return self._require_start()

    def require_end(self):
        return self._require_end()

    def require_center(self):
        return self._require_center()

    def require_width(self):
        return self._require_width()

    def get_n_dims(self):
        return self._n_dims

    def get_shape(self):
        return self._shape

    def convert_to_se(self, inplace=False):
        return self._convert_to_se(inplace=inplace)

    def convert_to_cw(self, inplace=False):
        return self._convert_to_cw(inplace=inplace)

    def convert_to_sw(self, inplace=False):
        return self._convert_to_cw(inplace=inplace)

    def get_bounds_se(self, copy=True):
        return self._get_bounds_se(copy=copy)

    def get_bounds_cw(self, copy=True):
        return self._get_bounds_cw(copy=copy)

    def get_bounds_sw(self, copy=True):
        _raise_not_overwritten(self)

    def get_center(self, copy=True):
        return self._get_center(copy=copy)

    def get_width(self, copy=True):
        return self._get_width(copy=copy)

    def get_start(self, copy=True):
        return self._get_start(copy=copy)

    def get_end(self, copy=True):
        return self._get_end(copy=copy)

    def scale_dimensions(self, factor_each_dim):
        factor_each_dim = self._reshape_input_for_application_to_each_bounds(factor_each_dim)
        return self._scale_dimensions(factor_each_dim)

    def scale_width(self, factor_each_dim):
        factor_each_dim = self._reshape_input_for_application_to_each_bounds(factor_each_dim)
        return self._scale_width(factor_each_dim)

    def add_offset(self, offset_each_dim):
        offset_each_dim = self._reshape_input_for_application_to_each_bounds(offset_each_dim)
        return self._add_offset(offset_each_dim)

    def apply_func_on_position(self, fn):
        return self._apply_func_on_position(fn)

    def apply_func_on_width(self, fn):
        return self._apply_func_on_width(fn)

    def copy(self):
        return self._copy()

    def delete(self, indices, axis=None):
        return self._delete(indices, axis=axis)

    def _require_se(self):
        _raise_not_overwritten(self)

    def _require_cw(self):
        _raise_not_overwritten(self)

    def _require_sw(self):
        _raise_not_overwritten(self)

    def _require_secw(self):
        _raise_not_overwritten(self)

    def _require_start(self):
        _raise_not_overwritten(self)

    def _require_end(self):
        _raise_not_overwritten(self)

    def _require_center(self):
        _raise_not_overwritten(self)

    def _require_width(self):
        _raise_not_overwritten(self)

    def _convert_to_se(self, inplace=False):
        _raise_not_overwritten(self)

    def _convert_to_cw(self, inplace=False):
        _raise_not_overwritten(self)

    def _convert_to_sw(self, inplace=False):
        _raise_not_overwritten(self)

    def _apply_func_on_position(self, fn):
        _raise_not_overwritten(self)

    def _apply_func_on_width(self, fn):
        _raise_not_overwritten(self)

    def _copy(self):
        _raise_not_overwritten(self)

    def _get_bounds_se(self, copy=True):
        _raise_not_overwritten(self)

    def _get_bounds_cw(self, copy=True):
        _raise_not_overwritten(self)

    def _get_bounds_sw(self, copy=True):
        _raise_not_overwritten(self)

    def _get_start(self, copy=True):
        _raise_not_overwritten(self)

    def _get_end(self, copy=True):
        _raise_not_overwritten(self)

    def _get_center(self, copy=True):
        _raise_not_overwritten(self)

    def _get_width(self, copy=True):
        _raise_not_overwritten(self)

    def _add_offset(self, offset_each_dim):
        _raise_not_overwritten(self)

    def _scale_dimensions(self, factor_each_dim):
        _raise_not_overwritten(self)

    def _scale_width(self, factor_each_dim):
        _raise_not_overwritten(self)

    def _delete(self, indices, axis=None):
        _raise_not_overwritten(self)

    def _getitem(self, key):
        _raise_not_overwritten(self)

    def _reshape_input_for_application_to_each_bounds(self, arr):
        arr = np.asarray(arr)
        if arr.shape == (self.get_n_dims(),):
            # Repeat the array for each bounds
            for n_elements in reversed(self.get_shape()):
                arr = np.repeat(arr[np.newaxis], n_elements, axis=0)
        elif arr.shape == self.get_shape()+(self.get_n_dims(),):
            # Shape already as required for calculations
            pass
        else:
            raise ValueError('Bad shape for input array. Expected {}, got {}.'.format(
                '{} or {}'.format((self.get_n_dims(),), self.get_shape()+(self.get_n_dims(),)),
                arr.shape
            ))

        return arr

    def _raise_not_possible_for_child(self):
        raise RuntimeError(
            'Method not possible for child (type = {}). '.format(type(self)) +
            'This error occurs usually, if the NdBounds class did not request the required attributes ' +
            'from the bound format handler.')

    def _pack_se(self, start, end):
        return NdBoundsSE(start=start, end=end)

    def _pack_cw(self, center, width):
        return NdBoundsCW(center=center, width=width)

    def _pack_sw(self, start, width):
        return NdBoundsSW(start=start, width=width)

    def _pack_sec(self, start, end, center):
        return NdBoundsSEC(start=start, end=end, center=center)

    def _pack_sew(self, start, end, width):
        return NdBoundsSEW(start=start, end=end, width=width)

    def _pack_scw(self, start, center, width):
        return NdBoundsSCW(start=start, center=center, width=width)

    def _pack_ecw(self, end, center, width):
        return NdBoundsECW(end=end, center=center, width=width)

    def _pack_secw(self, start, end, center, width):
        return NdBoundsSECW(start=start, end=end, center=center, width=width)


# Imports must be below the base class, since it will will be imported by subclasses
from .format_se import NdBoundsSE
from .format_cw import NdBoundsCW
from .format_sw import NdBoundsSW
from .format_sec import NdBoundsSEC
from .format_sew import NdBoundsSEW
from .format_scw import NdBoundsSCW
from .format_ecw import NdBoundsECW
from .format_secw import NdBoundsSECW


def _raise_not_overwritten(inst):
    raise RuntimeError('Method should be overwritten by child (type = {})'.format(type(inst)))



# Copyright (C) 2021 Matthias Nadig

import numpy as np


def _flatten_component(bounds_component):
    """ Flattens the array except for the last axis -> output shape is (None, n_dims) """
    if bounds_component.ndim == 1:
        # Singular bound
        return bounds_component
    else:
        return np.reshape(bounds_component, (-1, bounds_component.shape[-1]))


def _copy_if_required(arr, copy):
    """ Returns either the input array or a copy of it, depending on the copy-flag """
    if copy:
        return np.copy(arr)
    else:
        return arr



# Copyright (C) 2021 Matthias Nadig

import numpy as np

from .. import _conversion as conversion
from .. import _utils as utils

from .base import BoundFormatHandler


class NdBoundsSEW(BoundFormatHandler):
    """ Handler containing: start, end, width """

    def __init__(self, start, end, width):
        super().__init__(n_dims=start.shape[-1], shape=start.shape[:-1])
        self._start = start
        self._end = end
        self._width = width

    def _require_se(self):
        return self

    def _require_cw(self):
        center, width = conversion._convert_se_to_cw(self._start, self._end, inplace=False)
        return self._pack_secw(start=self._start, end=self._end, center=center, width=width)

    def _require_sw(self):
        return self

    def _require_secw(self):
        center, width = conversion._convert_se_to_cw(self._start, self._end, inplace=False)
        return self._pack_secw(start=self._start, end=self._end, center=center, width=width)

    def _require_start(self):
        return self

    def _require_end(self):
        return self

    def _require_center(self):
        center, width = conversion._convert_se_to_cw(self._start, self._end, inplace=False)
        return self._pack_secw(start=self._start, end=self._end, center=center, width=width)

    def _require_width(self):
        return self

    def _convert_to_se(self, inplace=False):
        return self._pack_se(start=self._start, end=self._end)

    def _convert_to_cw(self, inplace=False):
        center, width = conversion._convert_sw_to_cw(self._start, self._width, inplace=inplace)
        return self._pack_cw(center=center, width=width)

    def _convert_to_sw(self, inplace=False):
        return self._pack_sw(start=self._start, width=self._width)

    def _get_bounds_se(self, copy=True):
        return np.stack([self._start, self._end], axis=-1)

    def _get_bounds_cw(self, copy=True):
        self._raise_not_possible_for_child()

    def _get_bounds_sw(self, copy=True):
        return np.stack([self._start, self._width], axis=-1)

    def _get_start(self, copy=True):
        return utils._copy_if_required(self._start, copy=copy)

    def _get_end(self, copy=True):
        return utils._copy_if_required(self._end, copy=copy)

    def _get_center(self, copy=True):
        self._raise_not_possible_for_child()

    def _get_width(self, copy=True):
        return utils._copy_if_required(self._width, copy=copy)

    def _add_offset(self, offset_each_dim):
        center = np.subtract(self._end, self._start)
        np.add(center, offset_each_dim, out=center)
        return self._pack_cw(center=center, width=self._width)

    def _scale_dimensions(self, factor_each_dim):
        np.multiply(self._start, factor_each_dim, out=self._start)
        np.multiply(self._width, factor_each_dim, out=self._width)
        return self._pack_sw(start=self._start, width=self._width)

    def _scale_width(self, factor_each_dim):
        center = np.add(self._start, np.divide(self._width, 2))
        np.multiply(self._width, factor_each_dim, out=self._width)
        return self._pack_cw(center=center, width=self._width)

    def _apply_func_on_position(self, fn):
        center = np.add(self._start, np.divide(self._width, 2))
        fn(center, out=center)
        return self._pack_cw(center=center, width=self._width)

    def _apply_func_on_width(self, fn):
        center = np.add(self._start, np.divide(self._width, 2))
        fn(self._width, out=self._width)
        return self._pack_cw(center=center, width=self._width)

    def _copy(self):
        return self._pack_sew(start=np.copy(self._start), end=np.copy(self._end), width=np.copy(self._width))

    def _delete(self, indices, axis=None):
        if axis is None:
            start = utils._flatten_component(self._start)
            end = utils._flatten_component(self._end)
            width = utils._flatten_component(self._width)
            axis = 0
        elif axis > len(self.get_shape()):
            raise ValueError('Cannot delete elements along axis {} for bounds of shape {}'.format(
                axis, self.get_shape()))
        else:
            start = self._start
            end = self._end
            width = self._width
        start = np.delete(start, indices, axis=axis)
        end = np.delete(end, indices, axis=axis)
        width = np.delete(width, indices, axis=axis)

        return self._pack_sew(start=start, end=end, width=width)

    def _getitem(self, key):
        start = self._start[key]
        end = self._end[key]
        width = self._width[key]
        return self._pack_sew(start=start, end=end, width=width)



# Copyright (C) 2021 Matthias Nadig

import numpy as np


def _convert_se_to_cw(start, end, inplace=False):
    if inplace:
        center_prepared = start
        width_prepared = end
    else:
        center_prepared = np.empty(start.shape, dtype=start.dtype)
        width_prepared = np.empty(end.shape, dtype=end.dtype)

    width = np.subtract(end, start, out=width_prepared)
    center = np.add(start, np.multiply(0.5, width), out=center_prepared)

    return center, width


def _convert_cw_to_se(center, width, inplace=False):
    if inplace:
        start_prepared = center
        end_prepared = width
    else:
        start_prepared = np.empty(center.shape, dtype=center.dtype)
        end_prepared = np.empty(width.shape, dtype=width.dtype)

    width_half = np.multiply(0.5, width, out=end_prepared)
    start = np.subtract(center, width_half, out=start_prepared)
    end = np.add(center, width_half, out=end_prepared)

    return start, end


def _convert_se_to_sw(start, end, inplace=False):
    if inplace:
        width_prepared = end
    else:
        width_prepared = np.empty(end.shape, dtype=end.dtype)

    width = np.subtract(end, start, out=width_prepared)

    return start, width


def _convert_cw_to_sw(center, width, inplace=False):
    if inplace:
        start_prepared = center
    else:
        start_prepared = np.empty(center.shape, dtype=center.dtype)

    width_half = np.multiply(0.5, width)
    start = np.subtract(center, width_half, out=start_prepared)

    return start, width


def _convert_sw_to_se(start, width, inplace=False):
    if inplace:
        end_prepared = width
    else:
        end_prepared = np.empty(width.shape, dtype=width.dtype)

    end = np.add(start, width, out=end_prepared)

    return start, end


def _convert_sw_to_cw(start, width, inplace=False):
    if inplace:
        center_prepared = start
    else:
        center_prepared = np.empty(start.shape, dtype=start.dtype)

    width_half = np.multiply(0.5, width)
    center = np.add(start, width_half, out=center_prepared)

    return center, width


def _convert_array_se_to_cw(bounds):
    """
    Convert from (start, end) to (center, width)
    """

    bounds_new = np.empty(bounds.shape)
    bounds_new[..., 1] = np.subtract(bounds[..., 1], bounds[..., 0], out=bounds_new[..., 1])
    bounds_new[..., 0] = np.add(bounds[..., 0], np.multiply(0.5, bounds_new[..., 1]), out=bounds_new[..., 0])

    return bounds_new


def _convert_array_se_to_cw_inplace(bounds):
    """
    Convert from (start, end) to (center, width)
    CAVEAT: Inplace modification!
    """

    np.subtract(bounds[..., 1], bounds[..., 0], out=bounds[..., 1])
    np.add(bounds[..., 0], np.multiply(0.5, bounds[..., 1]), out=bounds[..., 0])

    return bounds


def _convert_array_se_to_sw(bounds):
    """
    Convert from (start, end) to (start, width)
    """
    return _convert_array_se_to_sw_inplace(np.copy(bounds))


def _convert_array_se_to_sw_inplace(bounds):
    """
    Convert from (start, end) to (start, width)
    CAVEAT: Inplace modification!
    """
    return np.subtract(bounds[..., 1], bounds[..., 0], out=bounds[..., 1])


def _convert_array_cw_to_se(bounds):
    """
    Convert from (center, width) to (start, end)
    """

    bounds_new = np.empty(bounds.shape)

    bounds_center = np.copy(bounds[..., 0])
    bounds_size_half = np.multiply(0.5, bounds[..., 1])

    bounds_new[..., 0] = np.subtract(bounds_center, bounds_size_half)
    bounds_new[..., 1] = np.add(bounds_center, bounds_size_half)

    return bounds_new


def _convert_array_cw_to_se_inplace(bounds):
    """
    Convert from (center, width) to (start, end)
    CAVEAT: Inplace modification!
    """

    bounds_center = np.copy(bounds[..., 0])
    bounds_size_half = np.multiply(0.5, bounds[..., 1], out=bounds[..., 1])

    np.subtract(bounds_center, bounds_size_half, out=bounds[..., 0])
    np.add(bounds_center, bounds_size_half, out=bounds[..., 1])

    return bounds


def _convert_array_cw_to_sw(bounds):
    """
    Convert from (center, width) to (start, width)
    """

    bounds_new = np.empty(bounds.shape)

    width_half = np.divide(bounds[..., 1], 2)

    bounds_new[..., 0] = np.subtract(bounds[..., 0], width_half, out=bounds_new[..., 0])
    bounds_new[..., 1] = bounds[..., 1]

    return bounds_new


def _convert_array_cw_to_sw_inplace(bounds):
    """
    Convert from (center, width) to (start, width)
    CAVEAT: Inplace modification!
    """

    width_half = np.divide(bounds[..., 1], 2)

    np.subtract(bounds[..., 0], width_half, out=bounds[..., 0])

    return bounds


def _convert_array_sw_to_se(bounds):
    """
    Convert from (start, width) to (start, end)
    """
    return _convert_array_sw_to_se_inplace(np.copy(bounds))


def _convert_array_sw_to_se_inplace(bounds):
    """
    Convert from (start, width) to (start, end)
    CAVEAT: Inplace modification!
    """
    bounds[..., 1] = np.add(bounds[..., 0], bounds[..., 1], out=bounds[..., 1])
    return bounds



# Copyright (C) 2021 Matthias Nadig


class NdBounds:
    """
    n-Dimensional Bounds
    """

    def __init__(self, bound_handler):
        self._bound_handler = bound_handler

    def __getitem__(self, key):
        return NdBounds(self._bound_handler[key])

    def __delitem__(self, key):
        # del self._bound_handler[key]
        raise NotImplementedError('To delete single bounds, use the delete(...)-method')

    def __setitem__(self, key, value):
        raise RuntimeError('Tried to modify bounds. Access by indexing is read-only.')

    def get_n_dims(self):
        """
        Returns the number of dimensions of the bounds
        (NOT OF THE NUMPY ARRAYS THEMSELVES)
        """
        return self._bound_handler.get_n_dims()

    def get_shape(self):
        """ Returns the shape of the Numpy array that contains the bounds """
        return self._bound_handler.get_shape()

    def get_bounds_se(self, copy=True):
        """ Returns the bounds in a Numpy array with format (start, end) """
        self._bound_handler = self._bound_handler.require_se()
        return self._bound_handler.get_bounds_se(copy=copy)

    def get_bounds_cw(self, copy=True):
        """ Returns the bounds in a Numpy array with format (center, width) """
        self._bound_handler = self._bound_handler.require_cw()
        return self._bound_handler.get_bounds_cw(copy=copy)

    def get_bounds_sw(self, copy=True):
        """ Returns the bounds in a Numpy array with format (start, width) """
        self._bound_handler = self._bound_handler.require_sw()
        return self._bound_handler.get_bounds_sw(copy=copy)

    def get_start(self, copy=True):
        """ Returns the start of the bounds """
        self._bound_handler = self._bound_handler.require_start()
        return self._bound_handler.get_start(copy=copy)

    def get_end(self, copy=True):
        """ Returns the end of the bounds """
        self._bound_handler = self._bound_handler.require_end()
        return self._bound_handler.get_end(copy=copy)

    def get_center(self, copy=True):
        """ Returns the center of the bounds """
        self._bound_handler = self._bound_handler.require_center()
        return self._bound_handler.get_center(copy=copy)

    def get_width(self, copy=True):
        """ Returns the width of the bounds """
        self._bound_handler = self._bound_handler.require_width()
        return self._bound_handler.get_width(copy=copy)

    def convert_to_se(self, inplace=False):
        """
        Makes sure, that the bounds are kept in format (start, end)
        Note 1: This method is meant for runtime improvements only. It might
        be useful, e.g., if the bounds are retrieved in SE format only and
        therefore an inplace conversion should be done, minimizing allocation
        of new arrays.
        Note 2: The inplace option refers to the underlying Numpy arrays,
        containing the bound information. Using the convert-method will
        always change the NdBounds instance inplace, by converting the
        bound handler.
        """
        self._bound_handler = self._bound_handler.convert_to_se(inplace=inplace)
        return self

    def convert_to_cw(self, inplace=False):
        """
        Makes sure, that the bounds are kept in format (center, width)
        Also see convert_to_se().
        """
        self._bound_handler = self._bound_handler.convert_to_cw(inplace=inplace)
        return self

    def require_cw(self):
        """ Makes sure that center and width are precomputed """
        self._bound_handler = self._bound_handler.require_cw()
        return self

    def require_se(self):
        """ Makes sure that start and end are precomputed """
        self._bound_handler = self._bound_handler.require_se()
        return self

    def require_sw(self):
        """ Makes sure that start and width are precomputed """
        self._bound_handler = self._bound_handler.require_sw()
        return self

    def require_secw(self):
        """ Makes sure that start, end, center and width are precomputed """
        self._bound_handler = self._bound_handler.require_secw()
        return self

    def require_start(self):
        """ Makes sure that start is precomputed """
        self._bound_handler = self._bound_handler.require_start()
        return self

    def require_end(self):
        """ Makes sure that end is precomputed """
        self._bound_handler = self._bound_handler.require_end()
        return self

    def require_center(self):
        """ Makes sure that center is precomputed """
        self._bound_handler = self._bound_handler.require_center()
        return self

    def require_width(self):
        """ Makes sure that width is precomputed """
        self._bound_handler = self._bound_handler.require_width()
        return self

    def scale_dimensions(self, factor_each_dim, inplace=False):
        """
        Scales the dimensions, the bounds refer to.
        Can be used e.g. when the signal the bounds - that are given in indices - refer to is downsampled.
        """
        if inplace:
            self._bound_handler = self._bound_handler.scale_dimensions(factor_each_dim)
            bounds_mod = self
        else:
            bounds_mod = NdBounds(self._bound_handler.copy())
            bounds_mod.scale_dimensions(factor_each_dim, inplace=True)
        return bounds_mod

    def scale_width(self, factor_each_dim, inplace=False):
        """
        Scales the width of the individual bounds.
        """
        self._bound_handler = self._bound_handler.require_width()
        if inplace:
            self._bound_handler = self._bound_handler.scale_width(factor_each_dim)
            bounds_mod = self
        else:
            bounds_mod = NdBounds(self._bound_handler.copy())
            bounds_mod.scale_width(factor_each_dim, inplace=True)
        return bounds_mod

    def add_offset(self, offset_each_dim, inplace=False):
        """
        Adds offset to the bounds.
        """
        if inplace:
            self._bound_handler = self._bound_handler.add_offset(offset_each_dim)
            bounds_mod = self
        else:
            bounds_mod = NdBounds(self._bound_handler.copy())
            bounds_mod.add_offset(offset_each_dim, inplace=True)
        return bounds_mod

    def apply_func_on_position(self, fn, inplace=False):
        """
        Applies a function to the position of the bounds, e.g. np.log(...).
        The function must have the following signature:
            fn(arr, out)
        The first argument "arr" is a Numpy array, on which the function will be applied,
        and the second one "out" must take an array of the same shape as "arr", into which
        the result will be written.
        """
        if inplace:
            self._bound_handler = self._bound_handler.apply_func_on_position(fn)
            bounds_mod = self
        else:
            bounds_mod = NdBounds(self._bound_handler.copy())
            bounds_mod.apply_func_on_position(fn, inplace=True)
        return bounds_mod

    def apply_func_on_width(self, fn, inplace=False):
        """
        Applies a function to the width of the bounds, also see apply_func_on_position().
        """
        self._bound_handler = self._bound_handler.require_width()
        if inplace:
            self._bound_handler = self._bound_handler.apply_func_on_width(fn)
            bounds_mod = self
        else:
            bounds_mod = NdBounds(self._bound_handler.copy())
            bounds_mod.apply_func_on_width(fn, inplace=True)
        return bounds_mod

    def copy(self):
        """ Returns a copy of itself """
        return NdBounds(self._bound_handler.copy())

    def delete(self, indices, axis=None):
        """ Deletes bounds at given indices """
        return NdBounds(self._bound_handler.delete(indices, axis=axis))



# Copyright (C) 2021 Matthias Nadig

from .base import BoundFormatHandler
from .format_se import NdBoundsSE
from .format_cw import NdBoundsCW
from .format_sw import NdBoundsSW
from .format_sec import NdBoundsSEC
from .format_sew import NdBoundsSEW
from .format_scw import NdBoundsSCW
from .format_ecw import NdBoundsECW
from .format_secw import NdBoundsSECW



# Copyright (C) 2021 Matthias Nadig

import numpy as np

from ._bounds import NdBounds
from . import _formats as formats


def as_bounds(start=None, end=None, center=None, width=None,
              n_dims=None, copy=True, assert_order=True):
    """
    Assembles NdBounds object.
    """
    if not (start is None or end is None or center is None or width is None):
        # SECW
        raise NotImplementedError()
    elif not (start is None or end is None) and center is None and width is None:
        # SE
        bounds = _assemble_from_start_end(start, end, copy=copy, n_dims=n_dims, assert_order=assert_order)
    elif start is None and end is None and not (center is None or width is None):
        # CW
        bounds = _assemble_from_center_width(center, width, copy=copy, n_dims=n_dims)
    elif not start is None and end is None and center is None and not width is None:
        # SW
        bounds = _assemble_from_start_width(start, width, copy=copy, n_dims=n_dims)
    elif not start is None and not end is None and center is None and not width is None:
        # SEW
        _raise_invalid_combination_of_inputs(start, end, center, width)
    elif not start is None and not end is None and not center is None and width is None:
        # SEC
        _raise_invalid_combination_of_inputs(start, end, center, width)
    elif not start is None and end is None and not center is None and not width is None:
        # SCW
        _raise_invalid_combination_of_inputs(start, end, center, width)
    elif start is None and not end is None and not center is None and not width is None:
        # ECW
        _raise_invalid_combination_of_inputs(start, end, center, width)
    else:
        _raise_invalid_combination_of_inputs(start, end, center, width)

    return bounds


def interpret_as_s1s2e1e2(bounds, n_dims=2, copy=True, assert_order=True):
    """
    Assembles NdBounds object from an array, which contains a common bounding box format.
    Shape must be (..., n_dims * 2).
    The dimensions are arbitrary. Per default, 2D is assumed for more explicitness.

    Basic format is:
        [start dim 1, start dim 2, ..., start dim n, end dim 1, end dim 2, ..., end dim n]
    """
    bounds = _assert_array(bounds, copy=copy)
    if n_dims is None:
        n_dims = bounds.shape[-1] // 2
    else:
        if not bounds.shape[-1] == n_dims * 2:
            raise ValueError('Shape must be (..., n_dims * 2) for {}D, got {}'.format(n_dims, bounds.shape))
    shape_bounds = bounds.shape[:-1]
    bounds = np.swapaxes(np.reshape(bounds, shape_bounds+(2, n_dims)), -2, -1)
    return _interpret_as_se(bounds, n_dims=None, copy=False, assert_order=assert_order)


def interpret_as_s1s2w1w2(bounds, n_dims=2, copy=True):
    """
    Assembles NdBounds object from an array, which contains a common bounding box format.
    Shape must be (..., n_dims * 2).
    The dimensions are arbitrary. Per default, 2D is assumed for more explicitness.

    Basic format is:
        [start dim 1, start dim 2, ..., start dim n, width dim 1, width dim 2, ..., width dim n]
    """
    bounds = _assert_array(bounds, copy=copy)
    if not bounds.shape[-1] == n_dims * 2:
        raise ValueError('Shape must be (..., n_dims * 2) for {}D, got {}'.format(n_dims, bounds.shape))
    shape_bounds = bounds.shape[:-1]
    bounds = np.swapaxes(np.reshape(bounds, shape_bounds+(2, n_dims)), -2, -1)
    return _assemble_from_start_width(bounds[..., 0], bounds[..., 1],
                                      n_dims=None, copy=False)


def interpret_as_se(bounds, n_dims=None, copy=True, assert_order=True):
    """
    Assembles NdBounds object from an array.
    Shape must be (..., n_dims, 2).
    Along the last axis, the array contains (start, end).

    Basic format is:
        [[start dim 1, end dim 1],
         [start dim 2, end dim 2],
         ...,
         [start dim n, end dim n]]
    """
    return _interpret_as_se(bounds, n_dims=n_dims, copy=copy, assert_order=assert_order)


def interpret_as_cw(bounds, n_dims=None, copy=True):
    """
    Assembles NdBounds object from an array.
    Shape must be (..., n_dims, 2).
    Along the last axis, the array contains (center, width).

    Basic format is:
        [[center dim 1, width dim 1],
         [center dim 2, width dim 2],
         ...,
         [center dim n, width dim n]]
    """
    return _interpret_as_cw(bounds, copy=copy, n_dims=n_dims)


def raise_if_not_ndbounds(bounds):
    """
    Helper, that asserts that input is an NdBounds object.
    """
    if not isinstance(bounds, NdBounds):
        raise TypeError('Expected input of type \'NdBounds\', instead got \'{}\''.format(type(bounds)))


def _assert_array(bounds, copy):
    if isinstance(bounds, (np.ndarray, list, tuple)):
        if copy:
            bounds = np.copy(bounds).astype(np.float)
        else:
            bounds = np.asarray(bounds, dtype=np.float)
    else:
        raise ValueError('Unexpected type of bounds array: {}'.format(bounds))

    return bounds


def _assert_shape_of_pair(bounds):
    # Assert bounds array to be of shape (..., n_dims, 2)
    if bounds.ndim < 3 or bounds.shape[-1] != 2:
        raise ValueError('Bounds array has bad shape: (..., n_dims, 2) != {}'.format(bounds.shape))


def _assert_shape_of_individuals(arr1, arr2):
    # Assert bounds arrays to be of same shape (..., n_dims)
    if arr1.ndim < 2 or arr1.shape != arr2.shape:
        raise ValueError('Bounds arrays must be of same shape (..., n_dims, 2), got {} and {} instead'.format(
            arr1.shape, arr2.shape))


def _assert_dimensionality_pair(bounds, n_dims=None):
    # Assert n_dims to be of certain value (dimensionality of bounds themselves, not the bounds array)
    if n_dims is not None:
        if bounds.shape[-2] != n_dims:
            raise ValueError(
                'Given bounds do not fulfil user requirements. ' +
                'Requested {}-dimensional bounds, got n_dims = {}'.format(n_dims, bounds.shape[-2]))


def _assert_dimensionality_individuals(*arrays, n_dims=None):
    # Assert n_dims to be of certain value (dimensionality of bounds themselves, not the bounds array)
    if n_dims is not None:
        n_arrays = len(arrays)
        n_dims_per_array = np.empty(n_arrays, dtype=np.int)
        for i in range(n_arrays):
            n_dims_per_array[i] = arrays[i].shape[-1]
        if (n_dims_per_array != n_dims).any():
            raise ValueError(
                'Given bounds do not fulfil user requirements. ' +
                'Requested {}-dimensional bounds, got n_dims = {}'.format(n_dims, tuple(n_dims_per_array)))


def _assert_order_pair(bounds):
    _assert_order_single(bounds[..., 0], bounds[..., 1])


def _assert_order_single(bounds_start, bounds_end):
    # Check order of start and end (only in case bounds of type start-end are required)
    is_wrong_order = np.less(np.subtract(bounds_end, bounds_start), 0)
    if is_wrong_order.any():
        str_count = '{} of {} bound(s)'.format(np.sum(is_wrong_order), is_wrong_order.size)
        raise ValueError('Bounds have wrong order: start > end for {}'.format(str_count))


def _interpret_as_se(bounds, n_dims=None, copy=True, assert_order=True):
    bounds = _assert_array(bounds, copy=copy)
    _assert_shape_of_pair(bounds)
    _assert_dimensionality_pair(bounds, n_dims=n_dims)
    if assert_order:
        _assert_order_pair(bounds)
    return NdBounds(formats.NdBoundsSE(start=bounds[..., 0], end=bounds[..., 1]))


def _assemble_from_start_end(start, end, n_dims=None, copy=True, assert_order=True):
    start = _assert_array(start, copy=copy)
    end = _assert_array(end, copy=copy)
    _assert_shape_of_individuals(start, end)
    _assert_dimensionality_individuals((start, end), n_dims=n_dims)
    if assert_order:
        _assert_order_single(start, end)
    return NdBounds(formats.NdBoundsSE(start=start, end=end))


def _interpret_as_cw(bounds, n_dims=None, copy=True):
    bounds = _assert_array(bounds, copy=copy)
    _assert_shape_of_pair(bounds)
    _assert_dimensionality_pair(bounds, n_dims=n_dims)
    return NdBounds(formats.NdBoundsCW(center=bounds[..., 0], width=bounds[..., 1]))


def _assemble_from_center_width(center, width, n_dims=None, copy=True):
    center = _assert_array(center, copy=copy)
    width = _assert_array(width, copy=copy)
    _assert_shape_of_individuals(center, width)
    _assert_dimensionality_individuals((center, width), n_dims=n_dims)
    return NdBounds(formats.NdBoundsCW(center=center, width=width))


def _assemble_from_start_width(start, width, n_dims=None, copy=True):
    start = _assert_array(start, copy=copy)
    width = _assert_array(width, copy=copy)
    _assert_shape_of_individuals(start, width)
    _assert_dimensionality_individuals((start, width), n_dims=n_dims)
    return NdBounds(formats.NdBoundsSW(start=start, width=width))


def _raise_invalid_combination_of_inputs(start, end, center, width):
    raise ValueError(
        'Combination of inputs bad or not supported:' +
        ('\n\t- {}' * 4).format(
            'start  is {}None'.format('' if start is None else 'not '),
            'end    is {}None'.format('' if end is None else 'not '),
            'center is {}None'.format('' if center is None else 'not '),
            'width  is {}None'.format('' if width is None else 'not '),
        ))



# Copyright (C) 2021 Matthias Nadig

from ._api import as_bounds
from ._api import interpret_as_se
from ._api import interpret_as_cw
from ._api import interpret_as_s1s2e1e2
from ._api import interpret_as_s1s2w1w2

from ._api import raise_if_not_ndbounds

__version__ = '3.0.0'
__author__ = 'Matthias Nadig'



# Copyright (C) 2021 Matthias Nadig

from setuptools import setup, find_packages


with open('README.md', 'r') as f:
    long_description = f.read()

setup(
    name='ndbounds',
    version='3.0.0',
    description='Toolbox for handling n-dimensional bounds',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Matthias Nadig',
    author_email='matthias.nadig@yahoo.com',
    license='MIT',
    packages=find_packages(),
    python_requires='>=3.4',
    install_requires=['numpy'],
)
